# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2024 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Records API."""

from invenio_records.dumpers import SearchDumper
from invenio_records.dumpers.relations import RelationDumperExt
from invenio_records.systemfields import ConstantField, DictField, ModelField
from invenio_records.systemfields.relations import MultiRelationsField
from invenio_records_resources.records.api import FileRecord, Record
from invenio_records_resources.records.systemfields import (
    FilesField,
    IndexField,
    PIDListRelation,
    PIDNestedListRelation,
    PIDRelation,
)
from invenio_vocabularies.contrib.affiliations.api import Affiliation
from invenio_vocabularies.contrib.awards.api import Award
from invenio_vocabularies.contrib.funders.api import Funder
from invenio_vocabularies.contrib.subjects.api import Subject
from invenio_vocabularies.records.api import Vocabulary
from invenio_vocabularies.records.systemfields.relations import CustomFieldsRelation

from invenio_communities.communities.records.systemfields.children import ChildrenField
from invenio_communities.communities.records.systemfields.is_verified import (
    IsVerifiedField,
)

from ..dumpers.featured import FeaturedDumperExt
from . import models
from .systemfields.access import CommunityAccessField
from .systemfields.deletion_status import CommunityDeletionStatusField
from .systemfields.parent_community import ParentCommunityField
from .systemfields.pidslug import PIDSlugField
from .systemfields.tombstone import TombstoneField


class CommunityFile(FileRecord):
    """Community file API."""

    model_cls = models.CommunityFileMetadata
    record_cls = None  # defined below


class Community(Record):
    """Community API."""

    id = ModelField()
    slug = ModelField()
    pid = PIDSlugField("id", "slug")
    parent = ParentCommunityField()
    children = ChildrenField()

    schema = ConstantField("$schema", "local://communities/communities-v1.0.0.json")

    model_cls = models.CommunityMetadata

    dumper = SearchDumper(
        extensions=[
            FeaturedDumperExt("featured"),
            RelationDumperExt("relations"),
        ]
    )

    index = IndexField("communities-communities-v2.0.0", search_alias="communities")

    access = CommunityAccessField()

    #: Custom fields system field.
    custom_fields = DictField(clear_none=True, create_if_missing=True)
    theme = DictField(clear_none=True)

    bucket_id = ModelField(dump=False)
    bucket = ModelField(dump=False)
    files = FilesField(
        store=False,
        file_cls=CommunityFile,
        # Don't delete, we'll manage in the service
        delete=False,
    )

    relations = MultiRelationsField(
        funding_award=PIDListRelation(
            "metadata.funding",
            relation_field="award",
            keys=[
                "title",
                "number",
                "identifiers",
                "acronym",
                "program",
                "subjects",
                "organizations",
            ],
            pid_field=Award.pid,
            cache_key="awards",
        ),
        funding_award_subjects=PIDNestedListRelation(
            "metadata.funding",
            relation_field="award.subjects",
            keys=["subject", "scheme", "props"],
            pid_field=Subject.pid,
            cache_key="subjects",
        ),
        funding_funder=PIDListRelation(
            "metadata.funding",
            relation_field="funder",
            keys=["name"],
            pid_field=Funder.pid,
            cache_key="funders",
        ),
        organizations=PIDListRelation(
            "metadata.organizations",
            keys=["name", "identifiers"],
            pid_field=Affiliation.pid,
            cache_key="affiliations",
        ),
        type=PIDRelation(
            "metadata.type",
            keys=["title"],
            pid_field=Vocabulary.pid.with_type_ctx("communitytypes"),
            cache_key="communitytypes",
        ),
        removal_reason=PIDRelation(
            "tombstone.removal_reason",
            keys=["title"],
            pid_field=Vocabulary.pid.with_type_ctx("removalreasons"),
            cache_key="removal_reason",
        ),
        custom=CustomFieldsRelation("COMMUNITIES_CUSTOM_FIELDS"),
    )

    is_verified = IsVerifiedField("is_verified", use_cache=True, index=True)

    deletion_status = CommunityDeletionStatusField()

    tombstone = TombstoneField()


CommunityFile.record_cls = Community



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
# Copyright (C) 2023 Graz University of Technology.
# Copyright (C) 2024 KTH Royal Institute of Technology.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Marshmallow fields."""

from invenio_i18n import lazy_gettext as _
from marshmallow import fields

from ...proxies import current_roles
from ...roles import Role


class RoleField(fields.Str):
    """Role field that serializes/deserializes/validates Role objects."""

    default_error_messages = {"invalid_role": _("Invalid role.")}

    def __init__(self, *args, **kwargs):
        """Field constructor.

        :param roles: A role registry. Defaults to the current roles.
        """
        self.roles = kwargs.pop("roles", current_roles)
        super().__init__(*args, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs):
        if value not in current_roles:
            raise self.make_error("invalid_role")
        return self.roles[value]

    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return value
        elif isinstance(value, Role):
            return value.name
        elif isinstance(value, str):
            if value in current_roles:
                return value
        raise RuntimeError(_("Not a valid role to serialize."))



#
# This file is part of Invenio.
# Copyright (C) 2023 CERN.
# Copyright (C) 2026 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Alter datetime columns to utc aware datetime columns."""

from invenio_db.utils import (
    update_table_columns_column_type_to_datetime,
    update_table_columns_column_type_to_utc_datetime,
)

# revision identifiers, used by Alembic.
revision = "c3feb26c7041"
down_revision = "72b37bb4119c"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    for table_name in [
        "communities_metadata",
        "communities_files",
        "communities_members",
        "communities_featured",
        "communities_archivedinvitations",
    ]:
        update_table_columns_column_type_to_utc_datetime(table_name, "created")
        update_table_columns_column_type_to_utc_datetime(table_name, "updated")
    update_table_columns_column_type_to_utc_datetime(
        "communities_featured", "start_date"
    )


def downgrade():
    """Downgrade database."""
    for table_name in [
        "communities_metadata",
        "communities_files",
        "communities_members",
        "communities_featured",
        "communities_archivedinvitations",
    ]:
        update_table_columns_column_type_to_datetime(table_name, "created")
        update_table_columns_column_type_to_datetime(table_name, "updated")
    update_table_columns_column_type_to_datetime("communities_featured", "start_date")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""Subcommunity resource configuration."""

from flask_resources import (
    HTTPJSONException,
    JSONDeserializer,
    JSONSerializer,
    RequestBodyParser,
    ResourceConfig,
    ResponseHandler,
    create_error_handler,
)
from invenio_records_resources.resources.records.headers import etag_headers
from invenio_records_resources.services.base.config import ConfiguratorMixin
from marshmallow import fields

from invenio_communities.communities.resources.args import (
    CommunitiesSearchRequestArgsSchema,
)

from ..services.errors import ParentChildrenNotAllowed

json_response_handler = ResponseHandler(JSONSerializer(), headers=etag_headers)


class SubCommunityResourceConfig(ConfiguratorMixin, ResourceConfig):
    """Subcommunity resource configuration."""

    # Blueprint configuration
    blueprint_name = "subcommunities"
    url_prefix = ""
    routes = {
        "join": "/communities/<pid_value>/actions/join-request",
    }
    request_view_args = {
        "pid_value": fields.UUID(),
    }

    # Request parsing
    request_read_args = {}
    request_extra_args = {}
    request_body_parsers = {"application/json": RequestBodyParser(JSONDeserializer())}
    default_content_type = "application/json"

    request_search_args = CommunitiesSearchRequestArgsSchema

    # Response handling
    response_handlers = {
        "application/json": json_response_handler,
        "application/vnd.inveniordm.v1+json": json_response_handler,
    }
    default_accept_mimetype = "application/json"

    # Error handling
    error_handlers = {
        # Parent community does not allow children
        ParentChildrenNotAllowed: create_error_handler(
            lambda e: HTTPJSONException(
                code=400,
                description=str(e),
            )
        ),
    }



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2024 CERN.
# Copyright (C) 2023-2025 Graz University of Technology.
# Copyright (C) 2024 KTH Royal Institute of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community schema."""

import re
from functools import partial
from uuid import UUID

from invenio_i18n import lazy_gettext as _
from invenio_records_resources.services.custom_fields import CustomFieldsSchema
from invenio_records_resources.services.records.schema import (
    BaseGhostSchema,
    BaseRecordSchema,
)
from invenio_vocabularies.contrib.affiliations.schema import (
    AffiliationRelationSchema as BaseAffiliationRelationSchema,
)
from invenio_vocabularies.contrib.awards.schema import FundingRelationSchema
from invenio_vocabularies.services.schema import (
    VocabularyRelationSchema as VocabularySchema,
)
from marshmallow import (
    EXCLUDE,
    Schema,
    ValidationError,
    fields,
    post_dump,
    post_load,
    pre_load,
    validate,
)
from marshmallow_utils.fields import (
    URL,
    ISODateString,
    NestedAttribute,
    SanitizedHTML,
    SanitizedUnicode,
)
from marshmallow_utils.permissions import FieldPermissionsMixin


def _not_blank(**kwargs):
    """Returns a non-blank validation rule."""
    max_ = kwargs.get("max", "")
    return validate.Length(
        error=_(
            "Field cannot be blank or longer than {max_} characters.".format(max_=max_)
        ),
        min=1,
        **kwargs,
    )


def no_longer_than(max, **kwargs):
    """Returns a character limit validation rule."""
    return validate.Length(
        error=_("Field cannot be longer than {max} characters.".format(max=max)),
        max=max,
        **kwargs,
    )


def is_not_uuid(value):
    """Make sure value is not a UUID."""
    try:
        UUID(value)
        raise ValidationError(
            _("The ID must not be an Universally Unique IDentifier (UUID).")
        )
    except (ValueError, TypeError):
        pass


class CommunityAccessSchema(Schema):
    """Community Access Schema."""

    visibility = fields.Str(
        validate=validate.OneOf(
            [
                "public",
                "restricted",
            ]
        )
    )
    members_visibility = fields.Str(
        validate=validate.OneOf(
            [
                "public",
                "restricted",
            ]
        )
    )
    member_policy = fields.Str(
        validate=validate.OneOf(
            [
                "open",
                "closed",
            ]
        )
    )
    record_policy = fields.Str(
        validate=validate.OneOf(
            [
                "open",
                "closed",
                "restricted",
            ]
        )
    )
    record_submission_policy = fields.Str(
        validate=validate.OneOf(
            [
                "open",
                "closed",
                "restricted",
            ]
        )
    )
    review_policy = fields.Str(
        validate=validate.OneOf(
            [
                "open",
                "closed",
                "members",
            ]
        )
    )


# TODO: Probably this should be the default behavior for all relations
class AffiliationRelationSchema(BaseAffiliationRelationSchema):
    """Relaxed affiliation relation schema."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE


class CommunityMetadataSchema(Schema):
    """Community metadata schema."""

    title = SanitizedUnicode(required=True, validate=_not_blank(max=250))
    description = SanitizedUnicode(validate=_not_blank(max=250))

    curation_policy = SanitizedHTML(validate=no_longer_than(max=50000))
    page = SanitizedHTML(validate=no_longer_than(max=50000))

    type = fields.Nested(VocabularySchema, metadata={"type": "communitytypes"})
    website = URL(validate=_not_blank())
    funding = fields.List(fields.Nested(FundingRelationSchema))
    organizations = fields.List(fields.Nested(AffiliationRelationSchema))

    # TODO: Add when general vocabularies are ready
    # domains = fields.List(fields.Str())


class AgentSchema(Schema):
    """An agent schema, using a string for the ID to allow the "system" user."""

    user = fields.String(required=True)


class RemovalReasonSchema(VocabularySchema):
    """Schema for the removal reason."""

    id = fields.String(required=True)


class TombstoneSchema(Schema):
    """Schema for the record's tombstone."""

    removal_reason = fields.Nested(RemovalReasonSchema)
    note = SanitizedUnicode()
    removed_by = fields.Nested(AgentSchema, dump_only=True)
    removal_date = ISODateString(dump_only=True)
    citation_text = SanitizedUnicode()
    is_visible = fields.Boolean()


class DeletionStatusSchema(Schema):
    """Schema for the record deletion status."""

    is_deleted = fields.Boolean(dump_only=True)
    status = fields.String(dump_only=True)


class CommunityThemeStyleSchema(Schema):
    """Community Theme configuration schema."""

    font = fields.Dict()
    primaryColor = fields.Str()
    secondaryColor = fields.Str()
    tertiaryColor = fields.Str()
    primaryTextColor = fields.Str()
    secondaryTextColor = fields.Str()
    tertiaryTextColor = fields.Str()
    mainHeaderBackgroundColor = fields.Str()


class CommunityThemeSchema(Schema):
    """Community theme schema."""

    style = fields.Nested(CommunityThemeStyleSchema)
    brand = fields.Str()
    enabled = fields.Boolean()


class ChildrenSchema(Schema):
    """Children schema."""

    allow = fields.Boolean()


class BaseCommunitySchema(BaseRecordSchema, FieldPermissionsMixin):
    """Base schema for the community metadata."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    field_dump_permissions = {
        # hide 'is_verified' behind a permission
        "is_verified": "moderate",
    }

    id = fields.String(dump_only=True)
    slug = SanitizedUnicode(
        required=True,
        validate=[
            _not_blank(max=100),
            validate.Regexp(
                r"^[-\w]+$",
                flags=re.ASCII,
                error=_(
                    "The identifier should contain only letters, numbers, or dashes."
                ),
            ),
            is_not_uuid,
        ],
    )
    metadata = NestedAttribute(CommunityMetadataSchema, required=True)
    access = NestedAttribute(CommunityAccessSchema, required=True)

    custom_fields = NestedAttribute(
        partial(CustomFieldsSchema, fields_var="COMMUNITIES_CUSTOM_FIELDS")
    )

    is_verified = fields.Boolean(dump_only=True)

    theme = fields.Nested(CommunityThemeSchema, allow_none=True)

    tombstone = fields.Nested(TombstoneSchema, dump_only=True)

    deletion_status = fields.Nested(DeletionStatusSchema, dump_only=True)

    children = NestedAttribute(ChildrenSchema)

    @post_dump
    def post_dump(self, data, many, **kwargs):
        """Hide tombstone info if the record isn't deleted and metadata if it is."""
        is_deleted = (data.get("deletion_status") or {}).get("is_deleted", False)
        tombstone_visible = (data.get("tombstone") or {}).get("is_visible", True)

        if data.get("custom_fields") is None:
            data.pop("custom_fields", None)

        if data.get("theme") is None:
            data.pop("theme", None)

        if not is_deleted or not tombstone_visible:
            data.pop("tombstone", None)

        return data


class CommunityParentSchema(BaseCommunitySchema):
    """Community parent schema."""


class CommunitySchema(BaseCommunitySchema):
    """Community schema."""

    parent = NestedAttribute(CommunityParentSchema, dump_only=True, allow_none=True)

    @post_dump
    def post_dump(self, data, many, **kwargs):
        """Hide parent field if it's not present."""
        data = super().post_dump(data, many, **kwargs)
        if data.get("parent") is None:
            data.pop("parent", None)
        return data

    @post_load(pass_original=True)
    def filter_parent_id(self, in_data, original_data, **kwargs):
        """Simply keep the parent id."""
        if "parent" in original_data:
            in_data["parent"] = (
                dict(id=original_data["parent"]["id"])
                if original_data["parent"]
                else None
            )
        return in_data

    @pre_load
    def initialize_custom_fields(self, data, **kwargs):
        """Ensure custom fields are initialized.

        We need to do that so that validation can take place in case a configured
        field is marked as required.
        """
        data.setdefault("custom_fields", {})
        return data

    @post_load
    def lowercase(self, in_data, **kwargs):
        """Ensure slug is lowercase."""
        in_data["slug"] = in_data["slug"].lower()
        return in_data


class CommunityFeaturedSchema(Schema):
    """Community Featured schema."""

    id = fields.Int(dump_only=True)
    start_date = fields.DateTime(
        required=True,
        metadata={
            "title": _("start date"),
            "description": _("Accepted format: YYYY-MM-DD hh:mm"),
            "placeholder": "YYYY-MM-DD hh:mm",
        },
    )


class CommunityGhostSchema(BaseGhostSchema):
    """Community ghost schema."""

    id = SanitizedUnicode(dump_only=True)
    metadata = fields.Constant(
        {
            "title": _("Deleted community"),
            "description": _("The community was deleted."),
        },
        dump_only=True,
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 CERN.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Notification related utils for notifications."""

from invenio_access.permissions import system_identity
from invenio_notifications.models import Recipient
from invenio_notifications.services.generators import RecipientGenerator
from invenio_records.dictutils import dict_lookup
from invenio_search.engine import dsl
from invenio_users_resources.proxies import current_users_service

from invenio_communities.proxies import current_communities


class CommunityMembersRecipient(RecipientGenerator):
    """Community member recipient generator for notifications."""

    def __init__(self, key, roles=None):
        """Ctor."""
        self.key = key
        self.roles = roles

    def __call__(self, notification, recipients: dict):
        """Fetch community and add members as recipients, based on roles."""
        community = dict_lookup(notification.context, self.key)
        filter_ = dsl.Q("terms", **{"role": self.roles}) if self.roles else None

        members = current_communities.service.members.scan(
            system_identity,
            community["id"],
            extra_filter=filter_,
        )

        user_ids = set()
        for m in members:
            # TODO: add support for groups
            if m["member"]["type"] != "user":
                continue
            user_ids.add(m["member"]["id"])

        # remove system_user_id if present
        user_ids.discard(system_identity.id)

        if not user_ids:
            return recipients

        filter_ = dsl.Q("terms", **{"id": list(user_ids)})
        users = current_users_service.scan(system_identity, extra_filter=filter_)
        for u in users:
            recipients[u["id"]] = Recipient(data=u)
        return recipients



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2023 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Communities search params module."""

from invenio_records_resources.services.records.params import ParamInterpreter

from invenio_communities.communities.records.systemfields.deletion_status import (
    CommunityDeletionStatusEnum,
)


class StatusParam(ParamInterpreter):
    """Evaluates the 'status' parameter."""

    def apply(self, identity, search, params):
        """Evaluate the status parameter on the search."""
        value = params.pop("status", None)
        if value is not None and value in [
            x.value for x in CommunityDeletionStatusEnum
        ]:
            search = search.filter("term", **{"deletion_status": value})
        return search


class IncludeDeletedCommunitiesParam(ParamInterpreter):
    """Evaluates the include_deleted parameter."""

    def apply(self, identity, search, params):
        """Evaluate the include_deleted parameter on the search."""
        value = params.pop("include_deleted", None)
        # Filter prevents from displaying deleted records on main site search
        # deleted records should appear only in admins panel
        if value is None:
            search = search.filter(
                "term",
                **{"deletion_status": CommunityDeletionStatusEnum.PUBLISHED.value},
            )
        return search



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2024 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community access system field."""

from enum import Enum, unique

from flask import current_app
from invenio_records.systemfields import SystemField


class AccessEnumMixin:
    """Mixin for enum functionalities."""

    @classmethod
    def validate(cls, level):
        """Validate a string against the enum values."""
        return cls(level) in cls

    def __str__(self):
        """Return its value."""
        return self.value


@unique
class VisibilityEnum(AccessEnumMixin, Enum):
    """Enum defining access visibility."""

    PUBLIC = "public"

    RESTRICTED = "restricted"


@unique
class MembersVisibilityEnum(AccessEnumMixin, Enum):
    """Enum defining members visibility."""

    PUBLIC = "public"

    RESTRICTED = "restricted"


@unique
class MemberPolicyEnum(AccessEnumMixin, Enum):
    """Enum defining member policies."""

    OPEN = "open"

    CLOSED = "closed"


@unique
class RecordSubmissionPolicyEnum(AccessEnumMixin, Enum):
    """Enum defining record submission policies."""

    OPEN = "open"

    CLOSED = "closed"


@unique
class ReviewPolicyEnum(AccessEnumMixin, Enum):
    """Enum defining review policies."""

    OPEN = "open"

    CLOSED = "closed"

    MEMBERS = "members"


class CommunityAccess:
    """Access management per community."""

    def __init__(
        self,
        visibility=None,
        members_visibility=None,
        member_policy=None,
        record_submission_policy=None,
        review_policy=None,
    ):
        """Create a new CommunityAccess object.

        :param visibility: The visibility level.
        """
        self.visibility = visibility or VisibilityEnum.PUBLIC
        self.members_visibility = members_visibility or MembersVisibilityEnum.PUBLIC
        self.member_policy = member_policy or MemberPolicyEnum.OPEN
        self.record_submission_policy = (
            record_submission_policy
            or current_app.config["COMMUNITIES_DEFAULT_RECORD_SUBMISSION_POLICY"]
        )
        self.review_policy = review_policy or ReviewPolicyEnum.CLOSED
        self.errors = []

    @classmethod
    def validate_visibility_level(cls, level):
        """Validate the visibility level."""
        return VisibilityEnum.validate(level)

    @classmethod
    def validate_members_visibility_level(cls, level):
        """Validate the visibility level."""
        return MembersVisibilityEnum.validate(level)

    @classmethod
    def validate_member_policy_level(cls, level):
        """Validate the member policy level."""
        return MemberPolicyEnum.validate(level)

    @classmethod
    def validate_record_submission_policy_level(cls, level):
        """Validate the record submission policy level."""
        return RecordSubmissionPolicyEnum.validate(level)

    @classmethod
    def validate_review_policy_level(cls, level):
        """Validate the review policy level."""
        return ReviewPolicyEnum.validate(level)

    @property
    def visibility(self):
        """Get the visibility level."""
        return self._visibility

    @visibility.setter
    def visibility(self, value):
        """Set the visibility level."""
        if not self.validate_visibility_level(value):
            raise ValueError(f"Unknown visibility level: {value}")
        self._visibility = value

    @property
    def members_visibility(self):
        """Get the members visibility level."""
        return self._members_visibility

    @members_visibility.setter
    def members_visibility(self, value):
        """Set the members visibility level."""
        if not self.validate_members_visibility_level(value):
            raise ValueError(f"Unknown members visibility level: {value}")
        self._members_visibility = value

    @property
    def visibility_is_public(self):
        """Return True when visibility is public."""
        return self.visibility == VisibilityEnum.PUBLIC.value

    @property
    def visibility_is_restricted(self):
        """Return True when visibility is restricted."""
        return self.visibility == VisibilityEnum.RESTRICTED.value

    @property
    def member_policy(self):
        """Get the member policy level."""
        return self._member_policy

    @member_policy.setter
    def member_policy(self, value):
        """Set the member policy level."""
        if not self.validate_member_policy_level(value):
            raise ValueError(f"Unknown member policy level: {value}")
        self._member_policy = value

    @property
    def record_submission_policy(self):
        """Get the record submission policy level."""
        return self._record_submission_policy

    @record_submission_policy.setter
    def record_submission_policy(self, value):
        """Set the record submission policy level."""
        if not self.validate_record_submission_policy_level(value):
            raise ValueError(f"Unknown record submission policy level: {value}")
        self._record_submission_policy = value

    @property
    def review_policy(self):
        """Get the review policy level."""
        return self._review_policy

    @review_policy.setter
    def review_policy(self, value):
        """Set the review policy level."""
        if not self.validate_review_policy_level(value):
            raise ValueError(f"Unknown review policy level: {value}")
        self._review_policy = value

    def dump(self):
        """Dump the field values as dictionary."""
        return {
            "visibility": str(self.visibility),
            "members_visibility": str(self.members_visibility),
            "member_policy": str(self.member_policy),
            "record_submission_policy": str(self.record_submission_policy),
            "review_policy": str(self.review_policy),
        }

    def refresh_from_dict(self, access_dict):
        """Re-initialize the Access object with the data in the access_dict."""
        new_access = self.from_dict(access_dict)
        self.visibility = new_access.visibility
        self.members_visibility = new_access.members_visibility
        self.member_policy = new_access.member_policy
        self.record_submission_policy = new_access.record_submission_policy
        self.review_policy = new_access.review_policy

    @classmethod
    def from_dict(
        cls,
        access_dict,
    ):
        """Create a new Access object from the specified 'access' property.

        The new ``CommunityAccess`` object will be populated with new instances
        from the configured classes.
        """
        errors = []

        access = cls(
            visibility=access_dict.get("visibility"),
            members_visibility=access_dict.get("members_visibility"),
            member_policy=access_dict.get("member_policy"),
            record_submission_policy=access_dict.get("record_submission_policy"),
            review_policy=access_dict.get("review_policy"),
        )
        access.errors = errors
        return access

    def __repr__(self):
        """Return repr(self)."""
        return (
            f"<{type(self).__name__} ("
            f"visibility: {str(self.visibility)}, "
            f"members_visibility: {str(self.members_visibility)}, "
            f"member_policy: {str(self.member_policy)}, "
            f"record_submission_policy: {str(self.record_submission_policy)}, "
            f"review_policy: {str(self.review_policy)})>"
        )


# TODO: Move to Invenio-Records-Resources and make reusable. Duplicated from
# Invenio-RDM-Records.
class CommunityAccessField(SystemField):
    """System field for managing record access."""

    access_obj_class = CommunityAccess

    def __init__(self, key="access", access_obj_class=None):
        """Create a new RecordAccessField instance."""
        self._access_obj_class = access_obj_class or self.access_obj_class
        super().__init__(key=key)

    def _from_dict(self, instance, data):
        """Allows to override behavior in subclass."""
        return

    def obj(self, instance):
        """Get the access object."""
        obj = self._get_cache(instance)
        if obj is not None:
            return obj

        data = self.get_dictkey(instance)
        if data:
            obj = self._access_obj_class.from_dict(data)
        else:
            obj = self._access_obj_class()

        self._set_cache(instance, obj)
        return obj

    def set_obj(self, record, obj):
        """Set the access object."""
        # We accept both dicts and access class objects.
        if isinstance(obj, dict):
            obj = self._access_obj_class.from_dict(obj)

        assert isinstance(obj, self._access_obj_class)

        # We do not dump the object until the pre_commit hook
        # I.e. record.access != record['access']
        self._set_cache(record, obj)

    def __get__(self, record, owner=None):
        """Get the record's access object."""
        if record is None:
            # access by class
            return self

        # access by object
        return self.obj(record)

    def __set__(self, record, obj):
        """Set the records access object."""
        self.set_obj(record, obj)

    def pre_commit(self, record):
        """Dump the configured values before the record is committed."""
        obj = self.obj(record)
        if obj is not None:
            # only set the 'access' property if one was present in the
            # first place -- this was a problem in the unit test:
            # tests/resources/test_resources.py:test_simple_flow
            record["access"] = obj.dump()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2021 CERN.
# Copyright (C) 2022-2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community database models."""

from datetime import datetime, timezone

from invenio_db import db
from invenio_files_rest.models import Bucket
from invenio_records.models import RecordMetadataBase
from invenio_records_resources.records import FileRecordModelMixin
from sqlalchemy.types import String
from sqlalchemy_utils.types import ChoiceType, UUIDType

from .systemfields.deletion_status import CommunityDeletionStatusEnum


class CommunityMetadata(db.Model, RecordMetadataBase):
    """Represent a community."""

    __tablename__ = "communities_metadata"

    slug = db.Column(String(255), unique=True, nullable=True)

    bucket_id = db.Column(UUIDType, db.ForeignKey(Bucket.id), index=True)
    bucket = db.relationship(Bucket)

    # The deletion status is stored in the model so that we can use it in SQL queries
    deletion_status = db.Column(
        ChoiceType(CommunityDeletionStatusEnum, impl=db.String(1)),
        nullable=False,
        default=CommunityDeletionStatusEnum.PUBLISHED.value,
    )


class CommunityFileMetadata(db.Model, RecordMetadataBase, FileRecordModelMixin):
    """File associated with a community."""

    __record_model_cls__ = CommunityMetadata

    __tablename__ = "communities_files"


class CommunityFeatured(db.Model, db.Timestamp):
    """Featured community entry."""

    __tablename__ = "communities_featured"

    id = db.Column(db.Integer, primary_key=True)
    community_id = db.Column(
        UUIDType, db.ForeignKey(CommunityMetadata.id), nullable=False
    )
    start_date = db.Column(
        db.UTCDateTime(),
        nullable=False,
        default=lambda: datetime.now(timezone.utc),
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""Subcommunities service."""

from datetime import datetime, timedelta, timezone

from invenio_i18n import gettext as _
from invenio_notifications.services.uow import NotificationOp
from invenio_records_resources.services.base import LinksTemplate, Service
from invenio_records_resources.services.records.schema import ServiceSchemaWrapper
from invenio_records_resources.services.uow import unit_of_work
from invenio_requests import current_request_type_registry
from invenio_requests.proxies import current_requests_service as requests_service
from werkzeug.local import LocalProxy

import invenio_communities.notifications.builders as notifications
from invenio_communities.proxies import current_communities, current_roles

from .errors import ParentChildrenNotAllowed
from .request import SubCommunityInvitationRequest

community_service = LocalProxy(lambda: current_communities.service)


class SubCommunityService(Service):
    """Subcommunities service."""

    @property
    def request_cls(self):
        """Request class."""
        return self.config.request_cls

    @property
    def schema(self):
        """Returns the data schema instance."""
        return ServiceSchemaWrapper(self, schema=self.config.schema)

    @property
    def links_item_tpl(self):
        """Item links template."""
        return LinksTemplate(
            self.config.links_item,
            context={
                "permission_policy_cls": self.config.permission_policy_cls,
            },
        )

    @property
    def expandable_fields(self):
        """Get expandable fields."""
        return []

    def _is_owner_of(self, identity, community):
        """Check if the identity provides community ownership."""
        return next(
            (
                x
                for x in identity.provides
                if x.method == "community"
                and x.value == community
                and x.role == current_roles.owner_role.name
            ),
            None,
        )

    @unit_of_work()
    def join(self, identity, id_, data, uow=None):
        """Request to join a subcommunity.

        Permissions are delegated to the communities and requests services. E.g. the
        request service will check if the user has the permission to act on behalf of the
        community.

        This method uses the unit of work pattern, therefore if any permission is
        denied, the transaction will be rolled back.
        """
        parent_community = community_service.record_cls.pid.resolve(id_)

        if not parent_community.children.allow:
            raise ParentChildrenNotAllowed(
                _("Parent community does not accept subcommunities.")
            )

        data_, errors = self.schema.load(
            data,
            context={"identity": identity, "community": parent_community},
            raise_errors=True,
        )

        community = None
        if "community_id" in data_:
            community = community_service.record_cls.pid.resolve(data_["community_id"])
        else:
            community = community_service.create(
                identity, data_["community"], uow=uow
            )._record

        self.require_permission(identity, "request_join", record=community)
        # Sender is the community
        creator = {"community": str(community.id)}

        # Receiver is the parent community
        receiver = {"community": str(parent_community.id)}

        # Topic is the community
        topic = {"community": str(community.id)}

        # Create and submit the request
        request_data = {
            "title": _('Inclusion of subcommunity "%(community)s"')
            % {"community": community.metadata["title"]}
        }
        request_payload = data.get("payload")
        if request_payload:
            request_data["payload"] = request_payload
        request = requests_service.create(
            identity, request_data, self.request_cls, receiver, creator, topic, uow=uow
        )
        uow.register(
            NotificationOp(
                notifications.SubCommunityCreate.build(
                    identity=identity, request=request._request
                )
            )
        )

        # Accept the request if the user is the owner of both communities
        is_owner_child = self._is_owner_of(identity, str(community.id))

        is_owner_parent = self._is_owner_of(identity, str(parent_community.id))
        if is_owner_child and is_owner_parent:
            request = requests_service.execute_action(
                identity, request.id, "accept", uow=uow
            )

        return request

        # TODO if the user is the owner of both communities, accept the request automatically

        # TODO add custom implementation for the result item
        # # We want wrap the request result item in a new one to modify the links
        # return self.result_item(
        #     self,
        #     identity,
        #     request._record,
        #     links_tpl=self.links_item_tpl,
        #     expandable_fields=self.expandable_fields,
        #     expand=True,
        # )

    @unit_of_work()
    def create_subcommunity_invitation_request(
        self,
        identity,
        parent_community_id,
        child_community_id,
        data,
        expires_at=None,
        uow=None,
    ):
        """Create and submit a SubCommunityInvitation request."""
        type_ = current_request_type_registry.lookup(
            SubCommunityInvitationRequest.type_id
        )
        parent_community = community_service.record_cls.pid.resolve(parent_community_id)
        child_community = community_service.record_cls.pid.resolve(child_community_id)

        if not expires_at:
            expires_at = datetime.now(timezone.utc) + timedelta(days=45)

        requests_service.create(
            identity=identity,
            data=data,
            request_type=type_,
            creator=parent_community,
            receiver=child_community,
            expires_at=expires_at,
            uow=uow,
        )



#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create member tables."""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql, postgresql
from sqlalchemy_utils import JSONType, UUIDType

# revision identifiers, used by Alembic.
revision = "fbe746957cfc"
down_revision = "f701a32e6fbe"
branch_labels = ()
depends_on = (
    "a14fa442680f",
    "2f63be7b7572",
)  # where the accounts_role and request_metadata table are created


def upgrade():
    """Upgrade database."""
    op.add_column(
        "communities_members", sa.Column("active", sa.Boolean(), nullable=False)
    )
    op.add_column(
        "communities_members", sa.Column("visible", sa.Boolean(), nullable=False)
    )
    op.add_column(
        "communities_members", sa.Column("group_id", sa.Integer(), nullable=True)
    )
    op.add_column(
        "communities_members", sa.Column("request_id", UUIDType(), nullable=True)
    )
    op.create_index(
        op.f("ix_communities_members_active"),
        "communities_members",
        ["active"],
        unique=False,
    )
    op.create_index(
        "ix_community_group",
        "communities_members",
        ["community_id", "group_id"],
        unique=True,
    )
    op.create_unique_constraint(
        op.f("uq_communities_members_request_id"), "communities_members", ["request_id"]
    )
    op.drop_constraint(
        "fk_communities_members_user_id_accounts_user",
        "communities_members",
        type_="foreignkey",
    )
    op.create_foreign_key(
        op.f("fk_communities_members_user_id_accounts_user"),
        "communities_members",
        "accounts_user",
        ["user_id"],
        ["id"],
        ondelete="RESTRICT",
    )
    op.create_foreign_key(
        op.f("fk_communities_members_group_id_accounts_role"),
        "communities_members",
        "accounts_role",
        ["group_id"],
        ["id"],
        ondelete="RESTRICT",
    )
    op.create_foreign_key(
        op.f("fk_communities_members_request_id_request_metadata"),
        "communities_members",
        "request_metadata",
        ["request_id"],
        ["id"],
        ondelete="SET NULL",
    )

    op.create_table(
        "communities_archivedinvitations",
        sa.Column(
            "created",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        ),
        sa.Column(
            "updated",
            sa.DateTime().with_variant(mysql.DATETIME(fsp=6), "mysql"),
            nullable=False,
        ),
        sa.Column(
            "json",
            sa.JSON()
            .with_variant(JSONType(), "mysql")
            .with_variant(
                postgresql.JSONB(none_as_null=True, astext_type=sa.Text()), "postgresql"
            )
            .with_variant(JSONType(), "sqlite"),
            nullable=True,
        ),
        sa.Column("version_id", sa.Integer(), nullable=False),
        sa.Column("id", UUIDType(), nullable=False),
        sa.Column("role", sa.String(length=50), nullable=False),
        sa.Column("visible", sa.Boolean(), nullable=False),
        sa.Column("active", sa.Boolean(), nullable=False),
        sa.Column("community_id", UUIDType(), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=True),
        sa.Column("group_id", sa.Integer(), nullable=True),
        sa.Column("request_id", UUIDType(), nullable=True),
        sa.ForeignKeyConstraint(
            ["community_id"],
            ["communities_metadata.id"],
            name=op.f(
                "fk_communities_archivedinvitations_community_id_communities_metadata"
            ),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["group_id"],
            ["accounts_role.id"],
            name=op.f("fk_communities_archivedinvitations_group_id_accounts_role"),
            ondelete="RESTRICT",
        ),
        sa.ForeignKeyConstraint(
            ["request_id"],
            ["request_metadata.id"],
            name=op.f("fk_communities_archivedinvitations_request_id_request_metadata"),
            ondelete="SET NULL",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["accounts_user.id"],
            name=op.f("fk_communities_archivedinvitations_user_id_accounts_user"),
            ondelete="RESTRICT",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_communities_archivedinvitations")),
        sa.UniqueConstraint(
            "request_id", name=op.f("uq_communities_archivedinvitations_request_id")
        ),
    )

    op.create_index(
        op.f("ix_communities_archivedinvitations_active"),
        "communities_archivedinvitations",
        ["active"],
        unique=False,
    )


def downgrade():
    """Downgrade database."""
    op.drop_constraint(
        op.f("fk_communities_members_request_id_request_metadata"),
        "communities_members",
        type_="foreignkey",
    )
    op.drop_constraint(
        op.f("fk_communities_members_group_id_accounts_role"),
        "communities_members",
        type_="foreignkey",
    )
    op.drop_constraint(
        op.f("fk_communities_members_user_id_accounts_user"),
        "communities_members",
        type_="foreignkey",
    )
    op.create_foreign_key(
        "fk_communities_members_user_id_accounts_user",
        "communities_members",
        "accounts_user",
        ["user_id"],
        ["id"],
        ondelete="CASCADE",
    )
    op.drop_constraint(
        op.f("uq_communities_members_request_id"), "communities_members", type_="unique"
    )
    op.drop_index("ix_community_group", table_name="communities_members")
    op.drop_index(
        op.f("ix_communities_members_active"), table_name="communities_members"
    )
    op.drop_column("communities_members", "request_id")
    op.drop_column("communities_members", "group_id")
    op.drop_column("communities_members", "visible")
    op.drop_column("communities_members", "active")

    op.drop_index(
        op.f("ix_communities_archivedinvitations_active"),
        table_name="communities_archivedinvitations",
    )
    op.drop_table("communities_archivedinvitations")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2024 CERN.
# Copyright (C) 2022 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community views."""

from .api import (
    blueprint,
    create_communities_api_blueprint,
    create_members_api_bp_from_app,
    create_subcommunities_api_blueprint,
)
from .ui import create_ui_blueprint

__all__ = (
    "blueprint",
    "create_communities_api_blueprint",
    "create_members_api_bp_from_app",
    "create_subcommunities_api_blueprint",
    "create_ui_blueprint",
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.
"""Subcommunities resource."""

from flask import g
from flask_resources import Resource, resource_requestctx, response_handler, route
from invenio_records_resources.resources.errors import ErrorHandlersMixin
from invenio_records_resources.resources.records.resource import (
    request_data,
    request_view_args,
)


class SubCommunityResource(ErrorHandlersMixin, Resource):
    """Subcommunity request resource."""

    def __init__(self, config, service):
        """Instantiate the resource with a given configuration and service."""
        super().__init__(config)
        self.service = service

    def create_url_rules(self):
        """Create the URL rules for the record resource."""
        routes = self.config.routes
        url_rules = [
            route("POST", routes["join"], self.join),
        ]
        return url_rules

    @request_view_args
    @response_handler()
    @request_data
    def join(self):
        """Join with a subcommunity."""
        result = self.service.join(
            identity=g.identity,
            id_=resource_requestctx.view_args["pid_value"],
            data=resource_requestctx.data,
        )
        return result.to_dict(), 201



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 CERN.
# Copyright (C) 2025 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Communities Resource Serializers."""

from flask_resources import BaseListSchema, JSONSerializer, MarshmallowSerializer

from invenio_communities.communities.resources.ui_schema import UICommunitySchema


class UICommunityJSONSerializer(MarshmallowSerializer):
    """UI Community JSON serializer."""

    def __init__(self):
        """Initialise Serializer."""
        super().__init__(
            format_serializer_cls=JSONSerializer,
            object_schema_cls=UICommunitySchema,
            list_schema_cls=BaseListSchema,
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 TU Wien.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Systemfield for managing tombstone information of a community."""

from datetime import datetime, timezone

from invenio_records.systemfields import SystemField
from invenio_requests.resolvers.registry import ResolverRegistry


class Tombstone:
    """A data class for managing tombstone information."""

    def __init__(self, data):
        """Constructor."""
        self.removal_reason = data.get("removal_reason")
        self.note = data.get("note")
        self.removed_by = data.get("removed_by")
        self.removal_date = data.get("removal_date")
        self.citation_text = data.get("citation_text")
        self.is_visible = data.get("is_visible", True)

    @property
    def removal_reason(self):
        """Get the removal reason."""
        return self._removal_reason

    @removal_reason.setter
    def removal_reason(self, value):
        """Set the removal reason, referencing the corresponding vocabulary.

        If the value is a string, it will be turned into a dictionary of the
        shape ``{"id": value}``.
        """
        # NOTE: the validity of the reference is checked
        #       in the record's relations systemfield
        if isinstance(value, str):
            value = {"id": value}

        self._removal_reason = value

    @property
    def note(self):
        """Get the public note attached to the tombstone."""
        return self._note

    @note.setter
    def note(self, value):
        """Set the public note attached to the tombstone."""
        self._note = value or ""

    @property
    def removed_by(self):
        """Get a reference to the entity who removed the record."""
        return self._removed_by

    @removed_by.setter
    def removed_by(self, value):
        """Set a reference to the entity who removed the record.

        If the value is a string or integer, it is assumed that the
        reference is a user and the value will be turned into a dictionary
        of the shape ``{"user": value}``.
        If the value is a referenceable entity (such as a ``User`` object),
        the ``ResolverRegistry`` will be used to create a reference dict.
        """
        if value is None:
            self._removed_by = None

        elif (ref := ResolverRegistry.reference_entity(value)) is not None:
            self._removed_by = ref

        elif isinstance(value, (int, str)):
            self._removed_by = {"user": str(value)}

        else:
            raise ValueError(f"Invalid value for 'tombstone.removed_by': {value}")

    @property
    def removal_date(self):
        """Get the removal date."""
        return self._removal_date

    @removal_date.setter
    def removal_date(self, value):
        """Set the removal date."""
        if value is None:
            value = datetime.now(timezone.utc)

        if isinstance(value, datetime):
            value = value.isoformat()

        self._removal_date = value

    @property
    def citation_text(self):
        """Get the citation text of the tombstoned community."""
        return self._citation_text

    @citation_text.setter
    def citation_text(self, value):
        """Set the citation text of the tombstoned community."""
        self._citation_text = value or ""

    @property
    def is_visible(self):
        """Check if the tombstone is publicly visible."""
        return self._is_visible

    @is_visible.setter
    def is_visible(self, value):
        """Set the public visibility of the tombstone."""
        self._is_visible = bool(value)

    @property
    def removed_by_proxy(self):
        """Resolve the entity proxy for ``self.removed_by``."""
        if self.removed_by is None:
            return None
        else:
            return ResolverRegistry.resolve_entity_proxy(self.removed_by)

    def dump(self):
        """Dump the tombstone into a dictionary."""
        data = {
            "note": self.note,
            "removed_by": self.removed_by,
            "removal_date": self.removal_date,
            "citation_text": self.citation_text,
            "is_visible": self.is_visible,
        }

        if self.removal_reason:
            data["removal_reason"] = self.removal_reason

        return data

    def __repr__(self):
        """Return repr(self)."""
        return repr(self.dump())


class TombstoneField(SystemField):
    """System field for accessing a community's deletion status."""

    #
    # Data descriptor methods (i.e. attribute access)
    #
    def __get__(self, record, owner=None):
        """Get the persistent identifier."""
        if record is None:
            return self  # returns the field itself.

        tombstone = self._get_cache(record)
        if tombstone is not None:
            return tombstone

        ts_dict = record.get("tombstone", None)
        tombstone = Tombstone(ts_dict) if ts_dict else None
        self._set_cache(record, tombstone)
        return tombstone

    def __set__(self, record, value):
        """Set a community's tombstone entry."""
        if value is None:
            tombstone = None
        elif isinstance(value, dict):
            tombstone = Tombstone(value)
        elif isinstance(value, Tombstone):
            tombstone = value
        else:
            raise ValueError(f"Invalid type for tombstone: {value}")

        self._set_cache(record, tombstone)

    def pre_commit(self, record):
        """Dump the configured tombstone before committing the community record."""
        tombstone = self._get_cache(record)
        if tombstone:
            record["tombstone"] = tombstone.dump()
        else:
            record.pop("tombstone", None)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Northwestern University.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Community base error."""

from flask_babel import ngettext
from invenio_i18n import gettext as _

from .utils import humanize_byte_size


class CommunityError(Exception):
    """Base exception for community errors."""


class CommunityFeaturedEntryDoesNotExistError(CommunityError):
    """The provided set spec does not exist."""

    def __init__(self, query_arguments):
        """Initialise error."""
        super().__init__(
            _(
                "A featured community entry with {q} does not exist.".format(
                    q=query_arguments
                )
            )
        )


class LogoNotFoundError(CommunityError):
    """The logo was not found."""

    def __init__(self):
        """Initialise error."""
        super().__init__(_("No logo exists for this community."))


class LogoSizeLimitError(CommunityError):
    """The provided logo size exceeds limit."""

    def __init__(self, limit, file_size):
        """Initialise error."""
        limit_value, limit_unit = humanize_byte_size(limit)
        file_size_value, file_size_unit = humanize_byte_size(file_size)
        super().__init__(
            _(
                "Logo size limit exceeded. Limit: {limit_value} {limit_unit} Given: {file_size_value} {file_size_unit}".format(
                    limit_value=limit_value,
                    limit_unit=limit_unit,
                    file_size_value=file_size_value,
                    file_size_unit=file_size_unit,
                )
            )
        )


class OpenRequestsForCommunityDeletionError(CommunityError):
    """There are open requests related to a specific community and cannot be deleted."""

    def __init__(self, requests):
        """Initialise error."""
        super().__init__(
            ngettext(
                "There is %(num)s request open for this community. Please, resolve it before deleting this community.",
                "There are %(num)s requests open for this community. Please, resolve all of them before deleting this community.",
                requests,
            )
        )


class CommunityDeletedError(CommunityError):
    """Error denoting that the community was deleted."""

    def __init__(self, community, result_item=None):
        """Constructor."""
        self.community = community
        self.result_item = result_item


class DeletionStatusError(CommunityError):
    """Indicator for the record being in the wrong deletion status for the action."""

    def __init__(self, community, expected_status):
        """Constructor."""
        self.community = community
        self.expected_status = expected_status


class SetDefaultCommunityError(CommunityError):
    """Record is not part of a community that is being set as default."""

    def __init__(self):
        """Initialise error."""
        super().__init__(
            _(
                "Cannot set community as the default. "
                "The record has not been added to the community."
            )
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2024 CERN.
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2023-2024 Graz University of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio communities extension."""

from flask_menu import current_menu
from flask_principal import identity_loaded
from invenio_accounts.signals import datastore_post_commit
from invenio_base.utils import load_or_import_from_config
from invenio_i18n import lazy_gettext as _
from invenio_records_resources.services import FileService

from invenio_communities.communities import (
    CommunityFileServiceConfig,
    CommunityResource,
    CommunityResourceConfig,
    CommunityService,
    CommunityServiceConfig,
)
from invenio_communities.members import (
    MemberResource,
    MemberResourceConfig,
    MemberService,
    MemberServiceConfig,
)
from invenio_communities.subcommunities import (
    SubCommunityResource,
    SubCommunityResourceConfig,
    SubCommunityService,
    SubCommunityServiceConfig,
)

from . import config
from .cache.cache import IdentityCache
from .roles import RoleRegistry
from .utils import load_community_needs, on_datastore_post_commit
from .views.ui import (
    _has_about_page_content,
    _has_curation_policy_page_content,
    _show_create_community_link,
)


class InvenioCommunities(object):
    """Invenio extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        app.extensions["invenio-communities"] = self

        self.init_services(app)
        self.init_resource(app)
        self.init_hooks(app)
        self.init_cache(app)

    def init_config(self, app):
        """Initialize configuration.

        Override configuration variables with the values in this package.
        """
        for k in dir(config):
            if k.startswith("COMMUNITIES_"):
                app.config.setdefault(k, getattr(config, k))

        self.roles_registry = RoleRegistry(app.config["COMMUNITIES_ROLES"])

    def init_services(self, app):
        """Initialize communities service."""
        # Services
        self.service = CommunityService(
            CommunityServiceConfig.build(app),
            files_service=FileService(CommunityFileServiceConfig.build(app)),
            members_service=MemberService(MemberServiceConfig.build(app)),
        )
        self.subcommunity_service = SubCommunityService(
            SubCommunityServiceConfig.build(app)
        )

    def init_resource(self, app):
        """Initialize communities resources."""
        # Resources
        self.communities_resource = CommunityResource(
            CommunityResourceConfig.build(app),
            self.service,
        )
        self.members_resource = MemberResource(
            MemberResourceConfig,
            self.service.members,
        )

        self.subcommunities_resource = SubCommunityResource(
            config=SubCommunityResourceConfig.build(app),
            service=self.subcommunity_service,
        )

    def init_hooks(self, app):
        """Initialize hooks."""
        datastore_post_commit.connect(on_datastore_post_commit)

        @identity_loaded.connect_via(app)
        def on_identity_loaded(_, identity):
            load_community_needs(identity)

    def cache_handler(self, app):
        """Return the cache handler."""
        handler_func = load_or_import_from_config(
            "COMMUNITIES_IDENTITIES_CACHE_HANDLER", app
        )
        handler = handler_func(app)
        assert isinstance(handler, IdentityCache)
        return handler

    def init_cache(self, app):
        """Initialize cache."""
        self.cache = self.cache_handler(app)


def api_finalize_app(app):
    """Finalize app."""
    init(app)


def finalize_app(app):
    """Finalize app."""
    init(app)
    register_menus(app)


def register_menus(app):
    """Register community menu items."""
    current_menu.submenu("main.communities").register(
        endpoint="invenio_communities.communities_frontpage",
        text=_("Communities"),
        order=1,
    )
    current_menu.submenu("plus.community").register(
        endpoint="invenio_communities.communities_new",
        text=_("New community"),
        order=2,
        visible_when=_show_create_community_link,
    )

    communities = current_menu.submenu("communities")

    communities.submenu("requests").register(
        endpoint="invenio_communities.communities_requests",
        text=_("Requests"),
        order=20,
        expected_args=["pid_value"],
        **{"icon": "inbox", "permissions": "can_search_requests"},
    )
    communities.submenu("members").register(
        endpoint="invenio_communities.members",
        text=_("Members"),
        order=30,
        expected_args=["pid_value"],
        **{"icon": "users", "permissions": "can_members_search_public"},
    )

    communities.submenu("settings").register(
        endpoint="invenio_communities.communities_settings",
        text=_("Settings"),
        order=40,
        expected_args=["pid_value"],
        **{"icon": "settings", "permissions": "can_update"},
    )

    communities.submenu("curation_policy").register(
        endpoint="invenio_communities.communities_curation_policy",
        text=_("Curation policy"),
        order=50,
        visible_when=_has_curation_policy_page_content,
        expected_args=["pid_value"],
        **{"icon": "balance scale", "permissions": "can_read"},
    )
    communities.submenu("about").register(
        endpoint="invenio_communities.communities_about",
        text=_("About"),
        order=60,
        visible_when=_has_about_page_content,
        expected_args=["pid_value"],
        **{"icon": "info", "permissions": "can_read"},
    )


def init(app):
    """Init app."""
    # Register services - cannot be done in extension because
    # Invenio-Records-Resources might not have been initialized.
    rr_ext = app.extensions["invenio-records-resources"]
    idx_ext = app.extensions["invenio-indexer"]
    ext = app.extensions["invenio-communities"]

    # services
    rr_ext.registry.register(ext.service, service_id="communities")
    rr_ext.registry.register(ext.service.members, service_id="members")

    # indexers
    idx_ext.registry.register(ext.service.indexer, indexer_id="communities")
    idx_ext.registry.register(ext.service.members.indexer, indexer_id="members")
    idx_ext.registry.register(
        ext.service.members.archive_indexer, indexer_id="archived-invitations"
    )

    # change notification handlers
    rr_ext.notification_registry.register("users", ext.service.on_relation_update)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 Graz University of Technology.
# Copyright (C) 2024 KTH Royal Institute of Technology.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Members Components."""

from flask_principal import Identity
from invenio_accounts.models import Role
from invenio_i18n import lazy_gettext as _
from invenio_records_resources.services.records.components import ServiceComponent

from invenio_communities.members.records.api import MemberMixin

from ...proxies import current_identities_cache
from ...utils import on_group_membership_change, on_user_membership_change


class CommunityMemberCachingComponent(ServiceComponent):
    """Service component for community member caching."""

    def _member_changed(self, member, community=None):
        """Call caching membership change function for user."""
        user_ids = []
        if isinstance(member, MemberMixin):
            if member.user_id:
                user_ids = [member.user_id]
            elif member.group_id:
                if not community:
                    raise TypeError(_("Community must be defined."))
                on_group_membership_change(str(community.id))
            else:
                return
        elif isinstance(member, dict):
            if member.get("type") == "user":
                user_ids = [member["id"]]
            elif member.get("type") == "group":
                if not community:
                    raise TypeError(_("Community must be defined."))
                on_group_membership_change(str(community.id))
            else:
                return
        else:
            raise TypeError(
                _("Member must be 'MemberMixin' or 'dict' but was %(type)s")
                % {"type": type(member)}
            )

        for user_id in user_ids:
            on_user_membership_change(Identity(user_id))

    def accept_invite(self, identity, record=None, data=None, **kwargs):
        """On accept invite."""
        self._member_changed(record)

    def members_add(self, identity, record=None, community=None, data=None, **kwargs):
        """On member add (only for groups)."""
        if record["type"] == "group":
            role = Role.query.filter_by(id=record["id"]).one_or_none()
            if role.is_managed:
                users = role.users.all()
                for user in users:
                    on_user_membership_change(Identity(user.id))
            else:
                current_identities_cache.flush()
        elif record["type"] == "user":
            self._member_changed(record)

    def members_delete(
        self, identity, record=None, community=None, data=None, **kwargs
    ):
        """On member delete."""
        self._member_changed(record, community=community)

    def members_update(
        self, identity, record=None, community=None, data=None, **kwargs
    ):
        """On member update."""
        self._member_changed(record, community=community)
